@@ -3,6 +3,7 @@
 /* For SIGWINCH */
 #include <skalibs/nonposix.h>
 
+#include <stdint.h>
 #include <unistd.h>
 #include <string.h>
 #include <strings.h>
@@ -16,7 +17,7 @@
 #include <skalibs/allreadwrite.h>
 #include <skalibs/bytestr.h>
 #include <skalibs/types.h>
-#include <skalibs/strerr2.h>
+#include <skalibs/strerr.h>
 #include <skalibs/tai.h>
 #include <skalibs/iopause.h>
 #include <skalibs/djbunix.h>
@@ -39,7 +40,7 @@ enum trans_e
 {
   V_TIMEOUT, V_CHLD, V_TERM, V_HUP, V_QUIT, V_INT,
   V_a, V_b, V_q, V_h, V_k, V_t, V_i, V_1, V_2, V_p, V_c, V_y, V_r,
-  V_o, V_d, V_u, V_x, V_O
+  V_o, V_d, V_u, V_D, V_U, V_x, V_O
 } ;
 
 typedef enum state_e state_t, *state_t_ref ;
@@ -52,16 +53,25 @@ enum state_e
   LASTFINISH
 } ;
 
+struct gflags_s
+{
+  uint8_t cont : 1 ;
+  uint8_t dying : 1 ;
+} gflags =
+{
+  .cont = 1,
+  .dying = 0
+} ;
+
 typedef void action_t (void) ;
 typedef action_t *action_t_ref ;
 
 static tain deadline ;
-static tain dontrespawnbefore = TAIN_EPOCH ;
+static tain nextstart = TAIN_ZERO ;
 static s6_svstatus_t status = S6_SVSTATUS_ZERO ;
 static state_t state = DOWN ;
-static int flagdying = 0 ;
-static int cont = 1 ;
 static int notifyfd = -1 ;
+static char const *servicename = 0 ;
 
 static inline void settimeout (int secs)
 {
@@ -117,10 +127,10 @@ static void set_down_and_ready (char const *s, unsigned int n)
   status.pid = 0 ;
   status.flagfinishing = 0 ;
   status.flagready = 1 ;
-  tain_wallclock_read(&status.readystamp) ;
   state = DOWN ;
-  if (tain_future(&dontrespawnbefore)) deadline = dontrespawnbefore ;
-  else tain_copynow(&deadline) ;
+  if (tai_sec(tain_secp(&nextstart))) deadline = nextstart ;
+  else tain_addsec_g(&deadline, 1) ;
+  tain_wallclock_read(&status.readystamp) ;
   announce() ;
   ftrigw_notifyb_nosig(S6_SUPERVISE_EVENTDIR, s, n) ;
 }
@@ -134,7 +144,7 @@ static void nop (void)
 
 static void bail (void)
 {
-  cont = 0 ;
+  gflags.cont = 0 ;
 }
 
 static void sigint (void)
@@ -155,6 +165,20 @@ static void closethem (void)
       strerr_warnwu2sys("open /dev/null for ", "writing") ;
 }
 
+static void adddown (void)
+{
+  if (!openwritenclose_unsafe("down", "", 0))
+    strerr_warnwu2sys("create ", "./down file") ;
+}
+
+static void deldown (void)
+{
+  int e = errno ;
+  if (unlink("down") == -1 && errno != ENOENT)
+    strerr_warnwu2sys("unlink ", "./down file") ;
+  errno = e ;
+}
+
 static void killa (void)
 {
   kill(status.pid, SIGALRM) ;
@@ -305,7 +329,7 @@ static void trystart (void)
   }
   else if (!pid)
   {
-    char const *cargv[2] = { "run", 0 } ;
+    char const *cargv[3] = { "run", servicename, 0 } ;
     ((char *)PROG)[strlen(PROG)] = ' ' ;
     selfpipe_finish() ;
     if (notifyp[0] >= 0) close(notifyp[0]) ;
@@ -358,11 +382,11 @@ static void trystart (void)
   fd_close(p[0]) ;
   notifyfd = notifyp[0] ;
   settimeout_infinite() ;
+  nextstart = tain_zero ;
   state = UP ;
   status.pid = pid ;
   status.flagready = 0 ;
   tain_wallclock_read(&status.stamp) ;
-  tain_addsec_g(&dontrespawnbefore, 1) ;
   announce() ;
   ftrigw_notifyb_nosig(S6_SUPERVISE_EVENTDIR, "u", 1) ;
   return ;
@@ -380,35 +404,46 @@ static void trystart (void)
   if (lfd >= 0) fd_close(lfd) ;
 }
 
+static void wantdown (void)
+{
+  status.flagwantup = 0 ;
+  announce() ;
+}
+
+static void wantup (void)
+{
+  status.flagwantup = 1 ;
+  announce() ;
+}
+
 static void downtimeout (void)
 {
   if (status.flagwantup) trystart() ;
   else settimeout_infinite() ;
 }
 
-static void down_O (void)
-{
-  status.flagwantup = 0 ;
-  announce() ;
-}
-
 static void down_o (void)
 {
-  down_O() ;
+  wantdown() ;
   trystart() ;
 }
 
 static void down_u (void)
 {
-  status.flagwantup = 1 ;
-  announce() ;
+  wantup() ;
   trystart() ;
 }
 
-static void down_d (void)
+static void down_D (void)
 {
-  status.flagwantup = 0 ;
-  announce() ;
+  adddown() ;
+  wantdown() ;
+}
+
+static void down_U (void)
+{
+  deldown() ;
+  down_u() ;
 }
 
 static int uplastup_z (void)
@@ -416,7 +451,7 @@ static int uplastup_z (void)
   status.wstat = (int)status.pid ;
   status.flagpaused = 0 ;
   status.flagready = 0 ;
-  flagdying = 0 ;
+  gflags.dying = 0 ;
   tain_wallclock_read(&status.stamp) ;
   if (notifyfd >= 0)
   {
@@ -455,7 +490,7 @@ static int uplastup_z (void)
   {
     char fmt0[UINT_FMT] ;
     char fmt1[UINT_FMT] ;
-    char *cargv[4] = { "finish", fmt0, fmt1, 0 } ;
+    char *cargv[5] = { "finish", fmt0, fmt1, (char *)servicename, 0 } ;
     selfpipe_finish() ;
     fmt0[uint_fmt(fmt0, WIFSIGNALED(status.wstat) ? 256 : WEXITSTATUS(status.wstat))] = 0 ;
     fmt1[uint_fmt(fmt1, WTERMSIG(status.wstat))] = 0 ;
@@ -490,7 +525,7 @@ static void lastup_z (void)
 
 static void uptimeout (void)
 {
-  if (flagdying)
+  if (gflags.dying)
   {
     killk() ;
     settimeout(5) ;
@@ -502,12 +537,6 @@ static void uptimeout (void)
   }
 }
 
-static void up_o (void)
-{
-  status.flagwantup = 0 ;
-  announce() ;
-}
-
 static void up_d (void)
 {
   tain tto ;
@@ -519,15 +548,21 @@ static void up_d (void)
   if (timeout && tain_from_millisecs(&tto, timeout))
   {
     tain_add_g(&deadline, &tto) ;
-    flagdying = 1 ;
+    gflags.dying = 1 ;
   }
   else settimeout_infinite() ;
 }
 
-static void up_u (void)
+static void up_D (void)
 {
-  status.flagwantup = 1 ;
-  announce() ;
+  adddown() ;
+  up_d() ;
+}
+
+static void up_U (void)
+{
+  deldown() ;
+  wantup() ;
 }
 
 static void up_x (void)
@@ -560,12 +595,6 @@ static void finish_z (void)
   else set_down_and_ready("D", 1) ;
 }
 
-static void finish_u (void)
-{
-  status.flagwantup = 1 ;
-  announce() ;
-}
-
 static void finish_x (void)
 {
   state = LASTFINISH ;
@@ -578,23 +607,23 @@ static void lastfinish_z (void)
   bail() ;
 }
 
-static action_t_ref const actions[5][24] =
+static action_t_ref const actions[5][26] =
 {
   { &downtimeout, &nop, &bail, &bail, &bail, &bail,
     &nop, &nop, &nop, &nop, &nop, &nop, &nop, &nop, &nop, &nop, &nop, &nop, &nop,
-    &down_o, &down_d, &down_u, &bail, &down_O },
+    &down_o, &wantdown, &down_u, &down_D, &down_U, &bail, &wantdown },
   { &uptimeout, &up_z, &up_term, &up_x, &bail, &sigint,
     &killa, &killb, &killq, &killh, &killk, &killt, &killi, &kill1, &kill2, &killp, &killc, &killy, &killr,
-    &up_o, &up_d, &up_u, &up_x, &up_o },
+    &wantdown, &up_d, &wantup, &up_D, &up_U, &up_x, &wantdown },
   { &finishtimeout, &finish_z, &finish_x, &finish_x, &bail, &sigint,
     &nop, &nop, &nop, &nop, &nop, &nop, &nop, &nop, &nop, &nop, &nop, &nop, &nop,
-    &up_o, &down_d, &finish_u, &finish_x, &up_o },
+    &wantdown, &wantdown, &wantup, &down_D, &up_U, &finish_x, &wantdown },
   { &uptimeout, &lastup_z, &up_d, &closethem, &bail, &sigint,
     &killa, &killb, &killq, &killh, &killk, &killt, &killi, &kill1, &kill2, &killp, &killc, &killy, &killr,
-    &up_o, &up_d, &nop, &closethem, &up_o },
+    &wantdown, &up_d, &wantup, &up_D, &up_U, &closethem, &wantdown },
   { &finishtimeout, &lastfinish_z, &nop, &closethem, &bail, &sigint,
     &nop, &nop, &nop, &nop, &nop, &nop, &nop, &nop, &nop, &nop, &nop, &nop, &nop,
-    &nop, &nop, &nop, &closethem, &nop }
+    &wantdown, &wantdown, &wantup, &down_D, &up_U, &closethem, &wantdown }
 } ;
 
 
@@ -612,6 +641,7 @@ static inline void handle_notifyfd (void)
     r = sanitize_read(fd_read(notifyfd, buf, 512)) ;
     if (r > 0 && memchr(buf, '\n', r))
     {
+      tain_addsec_g(&nextstart, 1) ;
       tain_wallclock_read(&status.readystamp) ;
       status.flagready = 1 ;
       announce() ;
@@ -676,8 +706,8 @@ static inline void handle_control (int fd)
     else if (!r) break ;
     else
     {
-      size_t pos = byte_chr("abqhkti12pcyroduxO", 18, c) ;
-      if (pos < 18) (*actions[state][V_a + pos])() ;
+      size_t pos = byte_chr("abqhkti12pcyroduDUxO", 20, c) ;
+      if (pos < 20) (*actions[state][V_a + pos])() ;
     }
   }
 }
@@ -765,6 +795,7 @@ int main (int argc, char const *const *argv)
   PROG = "s6-supervise" ;
   if (argc < 2) strerr_dieusage(100, USAGE) ;
   if (chdir(argv[1]) < 0) strerr_diefu2sys(111, "chdir to ", argv[1]) ;
+  servicename = argv[1] ;
   {
     size_t proglen = strlen(PROG) ;
     size_t namelen = strlen(argv[1]) ;
@@ -804,12 +835,12 @@ int main (int argc, char const *const *argv)
 
     tain_now_set_stopwatch_g() ;
     settimeout(0) ;
-    tain_wallclock_read(&status.stamp) ;
+    tain_copynow(&status.stamp) ;
     status.readystamp = status.stamp ;
     announce() ;
     ftrigw_notifyb_nosig(S6_SUPERVISE_EVENTDIR, "s", 1) ;
 
-    while (cont)
+    while (gflags.cont)
     {
       int r ;
       x[2].fd = notifyfd ;
